# Дистрибутиран Elixir

## Функции за статус на node

- Как да видим името на node-a си?

```elixir
node()
#=> :nonode@nohost
```


### Функции за статус на node

- Как да направим node-а си 'жив'?

```elixir
Node.alive?
#=> false

Node.start(:meddle, :shortnames)
#=> {:ok, #PID<0.115.0>}
node()
#=> :"meddle@meddlandr"
Node.alive?
#=> true
```


### Свързване на node-ове

- Списък на свързаните node-ове.

```elixir
Node.list()
#=> []
```


### Свързване на node-ове

- Свързване на node-ове.

```elixir
# @meddle
Node.connect(:ivan@meddlandr)
#=> true

Node.list()
#=> [:ivan@meddlandr]
```


### Свързване на node-ове

- Връзките между node-ове са симетрични и транзитивни.

```elixir
# @ivan
Node.list()
#=> [:meddle@meddlandr]

# @njichev
Node.connect(:ivan@meddlandr)
#=> true

Node.list()
#=> [:ivan@meddlandr, :meddle@meddlandr]
```


### Свързване на node-ове - Kubernetes

```elixir
config :libcluster,
  topologies: [
    k8s: [
      strategy: Elixir.Cluster.Strategy.Kubernetes,
      config: [
        mode: :dns,
        kubernetes_node_basename: "sanbase",
        kubernetes_selector: "app=sanbase",
        polling_interval: 10_000
      ]
    ]
  ]
```


### Проверка дали node е в мрежата

```elixir
# @ivan
Node.ping(:njichev)
#=> :pang
Node.ping(:njichev@meddlandr)
#=> :pong

Node.ping(:"slavi@127.0.0.1")
#=> [error] ** System NOT running to use fully qualified hostnames **
#=> ** Hostname 127.0.0.1 is illegal **
:pang
```


### Файлът *.hosts.erlang*

- Съдържа списък от познати node-ове, които биха могли да се намерят в мрежата.
- Файлът *.hosts.erlang*:

```
'192.168.3.10'.
'192.168.3.11'.
'192.168.3.12'.
```

- Функциата `:net_adm.names/0` ще листне всичко от там:

```elixir
:net_adm.names()
#=> {:ok, [{'b', 45325}, {'a', 33599}]}
```


### Създаване на процеси на отдалечени node-ове

- Основата на комуникацията между node-ове са процесите.
- `Node.spawn` и неговите версии:

```elixir
Node.spawn(:ivan@meddlandr, fn -> IO.puts(node()) end)
#=> #PID<13382.125.0>
#=> ivan@meddlandr
```


### Създаване на процеси на отдалечени node-ове

```elixir
# @slavi

defmodule DB do
  def tell_us_secret do
    IO.puts("Познавам всички и всеки!")
  end
end
```


### Създаване на процеси на отдалечени node-ове

```elixir
# @meddle
Node.spawn(:slavi@meddlandr, DB, :tell_us_secret, [])
#output: Познавам всички и всеки!
#=> #PID<9473.144.0>
```


### Създаване на процеси на отдалечени node-ове

```elixir
# @meddle
Node.spawn(:ivan@meddlandr, fn -> Process.sleep(60_000); IO.puts(node()) end)
#=> #PID<13382.126.0>
Process.group_leader
#=> #PID<0.66.0>

# @ivan
Process.info(pid(0, 129, 0), :group_leader)
#=> {:group_leader, #PID<11629.66.0>} # Забележете, че първият елемент на pid не е 0!
```


### Глобален регистър на процеси

- Къде ще се отпечата "Hey" и защо?

```elixir
# a@127.0.0.1 (ECIO)
device_pid = Process.whereis(ECIO.Device)
#=> #PID<0.234.0>
:global.register_name(ECIO.Device, device_pid)
#=> :yes

# b@127.0.0.1 (Asynch)

IO.puts(:global.whereis_name(ECIO.Device), "Hey")
#=> :ok
```


### Наблюдаване на Node-ове

- Можем да наблюдаваме node кога става offline.
- Същата функция с `false` ще спре наблюдението.
- Процесът извикал тази функция с `true` трябва да приема съобщението `{:nodedown, <nodename>}`.

```elixir
# @meddle
Node.monitor(:njichev@meddlandr, true)
#=> true

# Kill @njichev!
flush()
#=> {:nodedown, :njichev@meddlandr}
```


### Извикване на отдалечени функции

- `:rpc.call/4` - Извиква отдалечена функция и връща резултат:

```elixir
# @meddle

:rpc.call(:slavi@meddlandr, DB, :tell_us_secret, [])
#=> Познавам всички и всеки!
#=> :ok
```


### Извикване на отдалечени функции

- `:rpc.cast/4` - Извиква отдалечена функция асинхронно:

```elixir
# @meddle

:rpc.cast(:ivan@meddlandr, Enum, :map, [[1, 2, 3], fn x -> Process.sleep(10_000); IO.inspect(self()); x * x end])
#=> true
#=> #PID<0.144.0>
#=> #PID<0.144.0>
#=> #PID<0.144.0>

# @ivan, докато се принтира:
Process.alive?(pid("0.144.0"))
#=> true
```


### Извикване на отдалечени функции

- Функциите `async_call/4` и `yield/1`:

```elixir
# @ivan
pid = :rpc.async_call(:meddle@meddland,
                      Enum,
                      :sort,
                      [(1..2_000_000) |> Enum.shuffle])
#=> #PID<0.115.0>

# Това няма да забие текущия процес докато резултата е готов.

:rpc.yield(pid)
# Ще чака за резултат, или ако е готов ще го върне
```


### Извикване на отдалечени функции

- Функцията `nb_yield/4` (non blocking yield).

```elixir
# @meddle
pid = :rpc.async_call(:slavi@meddland, Process, :sleep, [50_000])
#=> #PID<0.108.0>

# По подразбиране timeout-ът е нула.
# Пробва и ако няма резултат - :timeout

:rpc.nb_yield(pid)
#=> :timeout
:rpc.nb_yield(pid, 1_000)
#=> :timeout

:rpc.nb_yield(pid, 50_000)
#=> {:value, :ok}
```


### Извикване на отдалечени функции

- Функциите `multicall/4` и `eval_everywhere`

```elixir
# @slavi
Node.list()
#=> [:meddle@meddland, :ivan@meddland]

:rpc.multicall([Node.self() | Node.list()], Node, :self, [])
#=> {[:slavi@meddland, :meddle@meddland, :ivan@meddland], []}
```


### EPMD и имена

```elixir
# @meddle
:net_adm.names()
#=> {:ok, [{'meddle', 38657}, {'ivan', 46597}, {'slavi', 33253}, {'andi', 41137}]}
```


### PIDs

```elixir
# @ivan

Node.spawn(:meddle@meddlandr, fn -> send(pid(0, 110, 0), "Hello from ivan!") end)
#=> #PID<11629.200.0>

# @meddle
flush()
#=> "Hello from ivan!"
#=> :ok
```


## External Term Format

```elixir
binary_pid = :erlang.term_to_binary(self())
#=> <<131, 88, 100, 0, 16, 109, 101, 100, 100, 108, 101, 64, 109, 101, 100, 100,
#=>  108, 97, 110, 100, 114, 0, 0, 0, 110, 0, 0, 0, 0, 100, 56, 62, 180>>
```


### External Term Format

- Винаги започват с 131
- Вторият байт е tag (в 1 байт) - за pid-ове : 88 (103 за по-стар формат)

```elixir
<<131, tag::binary-size(1), rest::binary >> = binary_pid
#=> <<131, 88, 100, 0, 16, 109, 101, 100, 100, 108, 101, 64, 109, 101, 100, 100,
#=>  108, 97, 110, 100, 114, 0, 0, 0, 110, 0, 0, 0, 0, 100, 56, 62, 180>>
```


### External Term Format

- За атомите, след tag байта следва закодирана стойността на атома.
- След името на node-a имаме 4 * 3 байта, които ще разгледаме след малко.

```elixir
name_size = byte_size(rest) - (4*3)
#=> 19

<<131, tag::binary-size(1), binary_atom::binary-size(name_size), rest::binary >> = binary_pid
#=> <<131, 88, 100, 0, 16, 109, 101, 100, 100, 108, 101, 64, 109, 101, 100, 100,
#=>  108, 97, 110, 100, 114, 0, 0, 0, 110, 0, 0, 0, 0, 100, 56, 62, 180>>
```


### External Term Format

- Името е атом, но във вложени term-ове нямаме *131* за начало.
- Атомите се кодират по няколко начина, *ATOM_EXT*, което е deprecated се ползва все още в pid-ове.
- Тагът на *ATOM_EXT* e 100, след това имаме 2 байта дължина на името и след това името:

```elixir
<<100, atom_len::integer-size(16), name::binary-size(atom_len)>> = binary_atom
#=> <<100, 0, 16, 109, 101, 100, 100, 108, 101, 64, 109, 101, 100, 100, 108, 97,
#=>  110, 100, 114>>

name
#=> "meddle@meddlandr"
```


### External Term Format

- Последните 12 байта са 3 integer-a
- id-то е номерът на pid-a (2-рото му число)
- serial e третото число на pid-a
- creation е дата в unix int, кога е бил създаден node-a.

```elixir
<<
  131,
  tag::binary-size(1),
  binary_atom::binary-size(name_size),
  id::integer-size(32),
  serial::integer-size(32),
  creation::integer-size(32)
>> = binary_pid
```